/* $RCSfile: StixCursor.h,v $
 * $Revision: 1.5 $ $Date: 2003/05/01 23:49:20 $
 * Auth: Dave Loffredo (loffredo@steptools.com)
 * 
 * Copyright (c) 1991-2003 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * This file is part of the STIX Library
 * 
 * This file may be distributed and/or modified under the terms of 
 * the GNU General Public License version 2 as published by the Free
 * Software Foundation and appearing in the file LICENSE.GPL included
 * with this file.
 * 
 * THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTY OF ANY KIND,
 * INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE.
 * 
 * STEP Tools Inc. also offers licenses for commercial use of this
 * software in non-GPL applications.  Contact info@steptools.com for
 * terms and pricing.
 * 
 * 		----------------------------------------
 *
 * Stix Cursors for traversing various types of one-to-many
 * relationships.
 */

#ifndef STIXCURSOR_H
#define STIXCURSOR_H

/* ======================================== */
/* Main Concept Cursors - These things allow easy traversal over the
 * internal indicies within the StixNC interface object.  Each one has
 * the following member functions:
 *
 * cursor_type (StixNC * nc);  -- Constructor, needs the StixNC object.
 *
 * object_type * next();     -- Call this repeatedly to traverse through
 * 				the objects.  It will return null when no
 *				more objects.
 *
 * void rewind();  	     -- resets the traversal to beginning.
 *
 * unsigned size();   	     -- returns the total number of objects present. 
 *				This may be calculated each time, so do not
 *			  	use this for loop guards.  Test for the null
 *				return from next() instead to check for no
 *				more objects.  
 *
 * The following cursors are available.  The C++ type of the return object
 * is also listed below.
 *
 * StixWorkpieceCursor	- product_definition
 * StixProjectCursor	- product
 * StixSetupCursor	- product
 * StixFeatureCursor	- shape_aspect
 * StixComponentCursor	- shape_aspect (profiles, paths, tapers, etc.)
 * StixExecutableCursor	- machining_process_executable
 * StixToolCursor	- machining_tool
 */

/* since this is very repetitive, use a macro */
#define DECLARE_STIXNC_CURSOR(cls,vartyp) \
class cls { 					\
 private: 					\
    StixNC *  m_nc;	/* index */		\
    unsigned  m_i; 	/* current */		\
						\
 public:					\
    cls (StixNC * nc);				\
						\
    /* returns null when no more objects */	\
    vartyp * next();				\
						\
    /* reset traversal to beginning */		\
    void rewind();				\
						\
    /* Not for loop guards.  Use null return 	\
     * from next() to signal no more objects. 	\
     */						\
    unsigned size();				\
}

DECLARE_STIXNC_CURSOR(StixWorkpieceCursor,product_definition);
DECLARE_STIXNC_CURSOR(StixProjectCursor,product);
DECLARE_STIXNC_CURSOR(StixSetupCursor,product);
DECLARE_STIXNC_CURSOR(StixFeatureCursor,shape_aspect);
DECLARE_STIXNC_CURSOR(StixComponentCursor,shape_aspect);
DECLARE_STIXNC_CURSOR(StixExecutableCursor,machining_process_executable);
DECLARE_STIXNC_CURSOR(StixToolCursor,machining_tool);


/* ========================================
 * Feature Boss Cursor - This traverses over the bosses associated
 * with certain types of features.  For convenience, this also has a
 * traverse() method to switch the feature that it looks at.
 */
class StixFeatureBossCursor {
 private:
    StixFeature *  m_idx;	/* feature */
    unsigned  	   m_i;		/* current sar */

 public:
    /* normally created by access functions */
    StixFeatureBossCursor (shape_aspect * feat);
    void traverse (shape_aspect * feat);

    /* returns null when no more objects */
    shape_aspect * next();

    /* reset traversal to beginning */
    void rewind();

    /* This is not for loop guards.  Use null return from next() to
     * signal no more objects.  This computes the size, it does not
     * just return a stored value. 
     */
    unsigned size();
};



/* ========================================
 * Workplan Element Cursor - This traverses over the substeps
 * associated with workplans.  For convenience, this has a traverse()
 * method to switch the workplan that it looks at.  This can traverse
 * using the next() function but since ordering is particularly
 * important, this also has a get() by index function so that you can
 * use either sequential or random access.
 * 
 * When update capability is added, this will also house the insert,
 * append, remove, and other update functions.
 */
class StixWorkplanElementCursor {
 private:
    machining_workplan *  m_wp;
    unsigned  	   	  m_i;	/* current elem */

 public:
    StixWorkplanElementCursor (machining_process_executable * wp);
    void traverse (machining_process_executable * wp);

    /* returns null when no more objects */
    machining_process_executable * next();

    /* returns null if out of bounds, cursor position not changed */
    machining_process_executable * get (unsigned i);

    /* returns the sequence position attribute of the underlying
     * sequential_method relationship, just for diagnostic purposes 
     */
    double get_sequence_position (unsigned i);

    /* reset traversal to beginning */
    void rewind();

    /* Usable for loop guards, but recommend caching the value */
    unsigned size();
};


/* ========================================
 * Workplan Nested Steps Cursor - This does a depth first search down
 * through the workplan elements and any nested workplans.  It returns
 * only workingsteps.  THIS IS A COMPUTED VALUE.
 *
 * Has a traverse() method to switch the workplan that it looks at.
 * This can traverse using next() or get() by index so you can access
 * sequentially or randomly. 
 */
class StixWorkplanAllNestedCursor {
 private:
    machining_workplan *  m_wp;
    unsigned  	   	  m_i;	/* current elem */

 public:
    StixWorkplanAllNestedCursor (machining_process_executable * wp);
    void traverse (machining_process_executable * wp);

    /* returns null when no more objects */
    machining_workingstep * next();

    /* returns null if out of bounds, cursor position not changed */
    machining_workingstep * get (unsigned i);

    /* reset traversal to beginning */
    void rewind();

    /* Usable for loop guards, but recommend caching the value */
    unsigned size();
};





/* ========================================
 * Workplan Unique Feature Cursor - This examines each workingstep in the
 * given worplan and any nested workplans.  It computes the set of unique
 * features associated with those steps.  THIS IS A COMPUTED VALUE.
 *
 * Has a traverse() method to switch the workplan that it looks at.
 * This can traverse using next() or get() by index so you can access
 * sequentially or randomly. 
 *
 * If this is passed a workingstep rather than a workplan, this cursor 
 * will do the right thing and return the one feature referenced by it.
 */
class StixWorkplanAllFeatureCursor {
 private:
    machining_workplan *  m_wp;
    machining_workingstep * m_ws;
    unsigned  	   	  m_i;	/* current elem */

 public:
    StixWorkplanAllFeatureCursor (machining_process_executable * wp);
    void traverse (machining_process_executable * wp);

    /* returns null when no more objects */
    shape_aspect * next();

    /* returns null if out of bounds, cursor position not changed */
    shape_aspect * get (unsigned i);

    /* reset traversal to beginning */
    void rewind();

    /* Usable for loop guards, but recommend caching the value */
    unsigned size();
};



/* ========================================
 * Workplan Unique Tool Cursor - This examines each workingstep in the
 * given worplan and any nested workplans.  It computes the set of unique
 * tools associated with those steps.  THIS IS A COMPUTED VALUE.
 *
 * Has a traverse() method to switch the workplan that it looks at.
 * This can traverse using next() or get() by index so you can access
 * sequentially or randomly. 
 *
 * If this is passed a workingstep rather than a workplan, this cursor 
 * will do the right thing and return the one tool referenced by it.
 */
class StixWorkplanAllToolCursor {
 private:
    machining_workplan *  m_wp;
    machining_workingstep * m_ws;
    unsigned  	   	  m_i;	/* current elem */

 public:
    StixWorkplanAllToolCursor (machining_process_executable * wp);
    void traverse (machining_process_executable * wp);

    /* returns null when no more objects */
    machining_tool * next();

    /* returns null if out of bounds, cursor position not changed */
    machining_tool * get (unsigned i);

    /* reset traversal to beginning */
    void rewind();

    /* Usable for loop guards, but recommend caching the value */
    unsigned size();
};



/* ========================================
 * Feature Face Cursor - This returns the advanced faces associated
 * with a particular feature.  If it is given an in-process feature,
 * it will return faces for the associated final feature instead, but
 * if none are found, it will return the faces for the in-process one.
 *
 * Has a traverse() method to switch the feature that it looks at.
 * This can traverse using next() or get() by index so you can access
 * sequentially or randomly.  The get() is not as efficient, so avoid
 * it if you can.
 *
 * Should this return face_surface rather than advanced_face??
 */
class StixFeatureFaceCursor {
 private:
    shape_aspect * 		m_sa;
    SetOfrepresentation_item * 	m_its;
    unsigned  	   	  	m_i;	/* current elem */

 public:
    StixFeatureFaceCursor (shape_aspect * sa);
    void traverse (shape_aspect * sa);

    /* Returns null when no more objects */
    advanced_face * next();

    /* Returns null if out of bounds, cursor position not changed,
     * Not very efficient, do not use if you can avoid it. 
     */
    advanced_face * get (unsigned i);

    /* reset traversal to beginning */
    void rewind();

    /* This is not for loop guards.  Use null return from next() to
     * signal no more objects.  This computes the size, it does not
     * just return a stored value. 
     */
    unsigned size();
};


#endif /* STIXCURSOR_H */
