/* $RCSfile: StixNC.cxx,v $
 * $Revision: 1.25 $ $Date: 2003/08/12 22:18:20 $
 * Auth: Dave Loffredo (loffredo@steptools.com)
 * 
 * Copyright (c) 1991-2003 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * This file is part of the STIX Library
 * 
 * This file may be distributed and/or modified under the terms of 
 * the GNU General Public License version 2 as published by the Free
 * Software Foundation and appearing in the file LICENSE.GPL included
 * with this file.
 * 
 * THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTY OF ANY KIND,
 * INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE.
 * 
 * STEP Tools Inc. also offers licenses for commercial use of this
 * software in non-GPL applications.  Contact info@steptools.com for
 * terms and pricing.
 * 
 * 		----------------------------------------
 */

#include <rose.h>
#include <RoseP21Parser.h>
#include <RoseP21Writer.h>

#include <integrated_cnc_schema.h>
#include <integrated_cnc_schema_ROSE_LOAD.h>
#include "stix_imp.h"

static int sequential_method_cmp (const void* a, const void* b);

int sequential_method_cmp (const void* a, const void* b)
{
    return (int) ((*(sequential_method **) a)-> sequence_position() -
		  (*(sequential_method **) b)-> sequence_position());
} 	

StixNC::StixNC() {
    m_design = 0;
    m_verbose = 1;
}

StixNC::~StixNC() 
{
    reset();
}


StixSchemaType StixNC::src_schema() {
    return StixSchemaManager::get_src_schema(design());
}

char * StixNC::src_schema_name() {
    return StixSchemaManager::get_schema_name(design()); 
}



/* public com methods */
int StixNC::load (const char * filename)
{
    /* Force linkers to bring in all C++ classes.  Without this, some
     * linkers leave out classes and we get "best-fit" messages when
     * matching EXPRESS to the C++ classes.
     */
    integrated_cnc_schema_force_load();

    reset();	/* clear all indicies */

    /* Preload the ap238 schema just to make sure everything 
     * is present.
     */
    RoseDesign * schema = stix_cnc_schema();
    if (!schema) return 1;

    /* register schema hook functions */
    RoseP21AddSchemaFn old_read = RoseP21Parser::add_schema_fn;
    RoseP21Parser::add_schema_fn = stix_schema_read; 

    /* load the test data */
    m_design = ROSE.findDesign(filename);

    /* restore schema hook and message severities */
    RoseP21Parser::add_schema_fn = old_read;
    stix_p21_restore_messages();

    if (!m_design) return 1;

    printf ("Building indicies\n");
    build_index();

    return 0;
}


void StixNC::reset()
{
    unsigned i, sz;

    /* Delete the index structures, any calculated values, and the
     * arrays that provide quick access to all of that stuff.  
     */

    /* index workpieces */
    for (i=0, sz=m_wkpcs.size(); i<sz; i++) {
	StixWorkpiece * idx = (StixWorkpiece*) m_wkpcs[i]; 
	delete idx;
    }    
    m_wkpcs.empty();
    
    /* index projects */
    for (i=0, sz=m_projects.size(); i<sz; i++) {
	StixProject * idx = (StixProject*) m_projects[i]; 
	delete idx;
    }    
    m_projects.empty();


    /* index setups */
    for (i=0, sz=m_setups.size(); i<sz; i++) {
	StixSetup * idx = (StixSetup*) m_setups[i]; 
	delete idx;
    }    
    m_setups.empty();

    /* index features */
    for (i=0, sz=m_features.size(); i<sz; i++) {
	StixFeature * idx = (StixFeature*) m_features[i]; 
	delete idx;
    }    
    m_features.empty();

    /* index feature components */
    for (i=0, sz=m_comps.size(); i<sz; i++) {
	StixComponent * idx = (StixComponent *) m_comps[i];
	delete idx;
    }    
    m_comps.empty();


    /* These lists held pointers to machining_process_executable
     * and machining_tool objects.  These will be deleted when the 
     * design goes, so we do not need to do anything more than just 
     * clean out the lists.
     */
    m_execs.empty();
    m_tools.empty();
    

    /* Get rid of the design -- The AIM data class extensions only
     * provide AIM links and do not need to be cleaned out.  Moving 
     * all of the AIM data to the trash is good enough.
     */
    if (m_design) {
	rose_move_to_trash (m_design);
	rose_empty_trash();
	m_design = 0;
    }
}


void StixNC::build_index() 
{
    /* identify the basic objects */
    classify_products();
    classify_aspects();
    classify_actions();
    classify_tolerances();

    /* add backpointers */
    link_pdrs();
    link_property_reps();
    link_action_reps();
    link_resource_reps();
    link_amrs();
    link_process_assocs();
    link_sars();

    /* computed things */
    workpiece_calc_extra();
    workstep_calc_extra();
    feature_calc_extra();
}


/* Product Categories -- a workpiece is none of the below */
#define CAT_PROJECT	0x001
#define CAT_SETUP	0x002
#define CAT_WORKPIECE	0x004
#define CAT_RAWPIECE	0x008

void StixNC::classify_products()
{
    RoseCursor objs;
    RoseObject * obj;
    unsigned i, sz;
    
    if (m_verbose) fprintf (stdout, "Classifying products and pdefs\n");

    /* categories for setup and project */
    objs.traverse (m_design);
    objs.domain (ROSE_DOMAIN(product_related_product_category));
    while (obj=objs.next()) {
	product_related_product_category * pc = 
	    ROSE_CAST(product_related_product_category, obj);

	/* watch for null strings */
	if (!pc->name() || !*pc->name()) continue;

	if (!strcmp (pc->name(), "machining project"))
	{
	    if (!pc->products()) continue;

	    /* tag as a project and make an index for them */
	    for (i=0, sz=pc->products()->size(); i<sz; i++) {
		product *p = pc->products()->get(i);
		StixProject * idx = new StixProject;

		p->m_catflags |= CAT_PROJECT;
		p->m_idx = idx;

		m_projects.append(idx);
		idx->m_product = p;
		idx->m_exid = m_projects.size() | EXID_PROJECT;
	    }
	}
	    
	if (!strcmp (pc->name(), "machining setup"))
	{
	    if (!pc->products()) continue;

	    /* tag as a setup and make an index for them */
	    for (i=0, sz=pc->products()->size(); i<sz; i++) {
		product *p = pc->products()->get(i);
		StixSetup * idx = new StixSetup;

		p->m_catflags |= CAT_SETUP;
		p->m_idx = idx;

		m_setups.append(idx);
		idx->m_product = p;
		idx->m_exid = m_setups.size() | EXID_SETUP;
	    }
	}
    }


    /* tag remaining products as workpieces */
    objs.domain (ROSE_DOMAIN(product_definition));
    while (obj=objs.next()) {
	product_definition * pd = ROSE_CAST(product_definition, obj);
	product * p = (pd->formation()? pd->formation()->of_product(): 0);
	
	if (!p) continue; /* should never happen */

	/* already tagged as a project or setup */
	if (p->m_catflags) continue;

	StixWorkpiece * idx = new StixWorkpiece;
	pd-> m_idx = idx;

	m_wkpcs.append(idx);
	idx->m_pdef = pd;
	idx->m_exid = m_wkpcs.size() | EXID_WORKPIECE;
	p->m_catflags |= CAT_WORKPIECE;
    }
}




void StixNC::classify_aspects()
{
    RoseCursor objs;
    RoseObject * obj;

    /* Like the product definitions, shape aspects are used for many
     * different purposes.
     */
    if (m_verbose) fprintf (stdout, "Classifying features\n");


    /* Traverse the instanced features and transition features and tag
     * them as features, and do a little bit of classification! 
     */
    objs.traverse (m_design);
    objs.domain (ROSE_DOMAIN(shape_aspect));
    while (obj=objs.next()) {
	shape_aspect * 	sa = ROSE_CAST(shape_aspect,obj);

	/* reduce types to an enum for simplicity */
	sa-> m_type = calc_aspect_base_type(sa);

	if (!sa-> m_type) continue;
	if (IS_FEATURE(sa)) {
	    StixFeature * idx = new StixFeature;
	    sa-> m_idx = idx;

	    m_features.append(idx);
	    idx->m_sa = sa;
	    idx->m_exid = m_features.size() | EXID_FEATURE;
	}
	else {
	    /* Some type of component, path or profile */
	    StixComponent * idx = new StixComponent;
	    sa-> m_idx = idx;

	    m_comps.append(idx);
	    idx->m_sa = sa;
	    idx->m_exid = m_comps.size() | EXID_COMPONENT;
	}
    }
}



void StixNC::classify_actions()
{
    RoseCursor objs;
    RoseObject * obj;
    unsigned i,sz;

    /* Go through and assign exids to the process things that can 
     * be referenced externally.
     */
    if (m_verbose) fprintf (stdout, "Classifying process things\n");

    /* Traverse the all of the process executables */
    objs.traverse (m_design);
    objs.domain (ROSE_DOMAIN(machining_process_executable));
    while (obj=objs.next()) {
	machining_process_executable * mpe = 
	    ROSE_CAST(machining_process_executable,obj);

	/* bypass these since they are not real workingsteps */
	if (mpe-> isa(ROSE_DOMAIN(machining_feature_process)) ||
	    mpe-> isa(ROSE_DOMAIN(machining_operator_instruction)))
	    continue;

	m_execs.append(mpe);
	mpe->m_exid = m_execs.size() | EXID_WORKSTEP;
    }


    /* Traverse the all of the machining tools */
    objs.domain (ROSE_DOMAIN(machining_tool));
    while (obj=objs.next()) {
	machining_tool * mt = ROSE_CAST(machining_tool,obj);

	m_tools.append(mt);
	mt->m_exid = m_tools.size() | EXID_TOOL;


	/* link tools back to operations */
	SetOfsupported_item * uses = mt-> usage();
	for (i=0, sz=uses->size(); i<sz; i++) {
	    RoseObject * it = get_select_object(uses->get(i));
	    if (!it-> isa (ROSE_DOMAIN(machining_operation)))
		continue;
	    
	    machining_operation * op = ROSE_CAST(machining_operation,it);
	    if (op-> m_tool) 
		fprintf (stdout, "WARN: operation #%d has multiple tools\n",
			 op-> entity_id());

	    op->m_tool = mt;
	}
    }
}



void StixNC::link_property_reps()
{
    RoseCursor objs;
    RoseObject *obj, *targ;

    /* Go through all of the property reps and chain them to their
     * owners.  Later on, we will go through them and pull out the 
     * shape reps.
     */

    if (m_verbose) fprintf (stdout, "Linking property reps\n");

    objs.traverse(m_design);
    objs.domain (ROSE_DOMAIN(property_definition_representation));
    while (obj=objs.next()) {
	property_definition_representation * pdr = 
	    ROSE_CAST(property_definition_representation, obj);

	/* separate shape reps from other properties */
	targ = get_property_owner(pdr);

	if (!targ) continue;  /* bad shape def */
	if (targ-> isa(ROSE_DOMAIN(product_definition)))
	{
	    product_definition * pd = 
		ROSE_CAST(product_definition,targ);

	    pdr-> m_next_pdr = pd->m_props;
	    pd->m_props = pdr;
	}
	if (targ-> isa(ROSE_DOMAIN(shape_aspect)))
	{
	    shape_aspect * sa = ROSE_CAST(shape_aspect,targ);

	    pdr-> m_next_pdr = sa->m_props;
	    sa->m_props = pdr;
	}
    }

    /* go through the list of all properties and extract the shape
     * reps for easy access.  We know these because they are hooked 
     * up by shape definition representation subtypes 
     */

    objs.domain (ROSE_DOMAIN(product_definition));
    while (obj=objs.next()) {
	StixWorkpiece * idx;
	product_definition * pd = ROSE_CAST(product_definition, obj);
	product * p = (pd->formation()? 
		       pd->formation()->of_product(): 0);

	property_definition_representation * pdr = pd-> m_props;
	    
	/* continue if not a workpiece */
	if (!p || !(p->m_catflags & CAT_WORKPIECE))
	    continue; /* not a workpiece */
	
	idx = get_workpiece_idx (pd);
	if (!idx) {
	    printf ("WARN: #%d related to a workpiece product, but not a workpiece.\n", 
		    obj-> entity_id());
	    continue; /* not a workpiece */
	}
	
	while (pdr) {
	    representation * rep = pdr-> used_representation();

	    if (!rep) {
		fprintf (stdout, "WARN: #%d has no representation\n",
			 pdr-> entity_id());
		pdr = pdr-> m_next_pdr;
		continue;
	    }

	    if (pdr-> isa(ROSE_DOMAIN(shape_definition_representation)))
	    {
		if (!rep-> isa(ROSE_DOMAIN(shape_representation))) {
		    fprintf (stdout, "WARN: #%d is not a shape rep\n",
			     rep-> entity_id());
		    pdr = pdr-> m_next_pdr;
		    continue;
		}

		shape_representation * srep = 
		    ROSE_CAST(shape_representation,rep);

		if (srep-> isa(ROSE_DOMAIN(shape_representation_with_parameters))) {
		    if (idx-> m_bbox)
			fprintf (stdout, "WARN: #%d has multiple bounding boxes\n",
				 pd-> entity_id());
		    idx-> m_bbox=srep;
		}
		else {
		    if (idx-> m_shape) 
			fprintf (stdout, "WARN: #%d has multiple shapes\n",
				 pd-> entity_id());
		    idx-> m_shape=srep;
		}
	    }
	    pdr = pdr-> m_next_pdr;
	}
    }

    objs.domain (ROSE_DOMAIN(shape_aspect));
    while (obj=objs.next()) {
	shape_representation ** imp;
	shape_representation ** exp;
	shape_aspect * sa = ROSE_CAST(shape_aspect, obj);
	property_definition_representation * pdr = sa-> m_props;

	/* continue if not a feature or component */
	if (!sa-> m_type) continue;

	/* shape aspects can also have shape reps with parameters
	 * associated with them.  Is it possible to have two?  one
	 * for the explicit and one for the implicit?
	 */

	if (IS_FEATURE(sa)) {
	    StixFeature * idx = get_feature_idx (sa);
	    imp = &(idx->m_implicit);
	    exp = &(idx->m_explicit);
	}
	else {
	    StixComponent * idx = get_component_idx (sa);
	    imp = &(idx->m_implicit);
	    exp = &(idx->m_explicit);
	}

	/* find shape reps with params for features and components */
	while (pdr) {
	    representation * rep = pdr-> used_representation();

	    if (!rep) {
		fprintf (stdout, "WARN: #%d has no representation\n",
			 pdr-> entity_id());
		pdr = pdr-> m_next_pdr;
		continue;
	    }

	    /* The shape_rep_w_params that holds the feature placement
	     * should be related using shape_definition_representation.  
	     * currently it is only related with a plain property_def_rep
	     */

	    int f = IS_FEATURE(sa);
	    if (rep-> isa(ROSE_DOMAIN(shape_representation_with_parameters))) 
	    {
  		if (!pdr-> isa(ROSE_DOMAIN(shape_definition_representation)))
  		    fprintf (stdout, "WARN: %s #%d implicit shape as normal property\n",
  			     (f? "feature": "component"), sa-> entity_id());


		if (*imp) fprintf (stdout, "WARN: #%d has multiple implicit shape reps\n",
				   sa-> entity_id());

		*imp= ROSE_CAST(shape_representation,rep);
	    }


	    else if (pdr-> isa(ROSE_DOMAIN(shape_definition_representation)))
	    {
		if (!rep-> isa(ROSE_DOMAIN(shape_representation))) {
		    fprintf (stdout, "WARN: #%d is not a shape rep\n",
			     rep-> entity_id());
		    pdr = pdr-> m_next_pdr;
		    continue;
		}

		if (*exp) fprintf (stdout, "WARN: #%d has multiple explicit shape reps\n",
				   sa-> entity_id());

		*exp= ROSE_CAST(shape_representation,rep);
	    }
	    pdr = pdr-> m_next_pdr;
	}
    }
}





void StixNC::link_action_reps()
{
    RoseCursor objs;
    RoseObject *obj, *targ;
    action_method * am;

    /* Go through all of the property reps and chain them to their
     * owners.  Later on, we will go through them and pull out the 
     * shape reps.
     */

    if (m_verbose) fprintf (stdout, "Linking action property reps\n");

    objs.traverse(m_design);
    objs.domain (ROSE_DOMAIN(action_property_representation));
    while (obj=objs.next()) {
	action_property_representation * apr = 
	    ROSE_CAST(action_property_representation, obj);

	targ = (apr->property()?
		get_select_object (apr->property()->definition()): 0);

	if (!targ) continue;  /* bad definition */
	if (targ-> isa(ROSE_DOMAIN(action_method)))
	{
	    am = ROSE_CAST(action_method,targ);
	    apr-> m_next_apr = am->m_props;
	    am->m_props = apr;
	}
    }
}



void StixNC::link_resource_reps()
{
    RoseCursor objs;
    RoseObject *obj, *targ;
    action_resource * ar;

    /* Go through all of the resource property reps and chain them to
     * their owners.  
     */

    if (m_verbose) fprintf (stdout, "Linking resource property reps\n");

    objs.traverse(m_design);
    objs.domain (ROSE_DOMAIN(resource_property_representation));
    while (obj=objs.next()) {
	resource_property_representation * rpr = 
	    ROSE_CAST(resource_property_representation, obj);

	targ = (rpr->property()?
		get_select_object (rpr->property()->resource()): 0);

	if (!targ) continue;  /* bad definition */
	if (targ-> isa(ROSE_DOMAIN(action_resource)))
	{
	    ar = ROSE_CAST(action_resource,targ);
	    rpr-> m_next_rpr = ar->m_props;
	    ar->m_props = rpr;
	}
    }
}



void link_amr_workplan (
    action_method_relationship * amr,
    machining_workplan * wp
    )
{
    action_method * ed = amr-> related_method();

    /* eventually key this off name == "workplan element" */
    if (amr-> isa (ROSE_DOMAIN(sequential_method))) 
    {
	sequential_method * sm; 
	machining_process_executable * mpe;
	machining_process_executable * old;

	if (!ed-> isa (ROSE_DOMAIN(machining_process_executable)))
	{
	    return;  /* not element of a workplan */
	}
	sm = ROSE_CAST(sequential_method, amr);
	wp-> m_steps.append (sm);

	mpe = ROSE_CAST(machining_process_executable, ed);
	old = mpe-> m_parent;
	if (old && 1){ //m_verbose) {
	    fprintf (stdout, "WARN: executable #%d has multiple parents\n",
		     mpe-> entity_id());
	}
	mpe-> m_parent = wp;
    }

    /* no other amrs involve workplan right now */
}


void link_amr_workingstep (
    action_method_relationship * amr,
    machining_workingstep * ws
    )
{
    char * name = amr-> name();
    action_method * ed = amr-> related_method();

    if (!name || !*name) {
	if (1) //m_verbose) 
	    fprintf (stdout, "WARN: amr #%d has no name\n",
		     amr-> entity_id());
	return;
    }

    if (!strcmp (name, "operation"))
    {
	if (!ed-> isa (ROSE_DOMAIN(machining_operation)))
	{
	    if (1) //m_verbose) 
		fprintf (stdout, "ERR: amr #%d bad operation\n",
			 amr-> entity_id());
	    return; 
	}
	if (ws-> m_op && 1){ //m_verbose) {
	    fprintf (stdout, "WARN: operation #%d has multiple technologies\n",
		     ws-> entity_id());
	}
	ws-> m_op = ROSE_CAST(machining_operation, ed);
    }
    /* unrecognized amr */
}



void link_amr_feature (
    action_method_relationship * amr,
    machining_feature_process * fp
    )
{
    machining_workingstep * ws;
    machining_feature_process ** mfp;

    char * name = amr-> name();
    action_method * ed = amr-> related_method();

    if (!name || !*name) {
	if (1) //m_verbose) 
	    fprintf (stdout, "WARN: amr #%d has no name\n",
		     amr-> entity_id());
	return;
    }

    if (!strcmp (name, "final feature") ||
	!strcmp (name, "in-process feature"))
    {
	if (!ed-> isa (ROSE_DOMAIN(machining_workingstep)))
	{
	    if (1) //m_verbose) 
		fprintf (stdout, "ERR: amr #%d bad $s assignment\n",
			 amr-> entity_id(), name);
	    return; 
	}

	/* cache amr in feature process */
	fp->m_steps.append(amr);

	/* link back from the workingstep */
	ws = ROSE_CAST(machining_workingstep,ed);
	mfp = (!strcmp (name, "final feature")?
	       &(ws->m_final):
	       &(ws->m_inproc));


	if (*mfp && 1){ //m_verbose) {
	    fprintf (stdout, "WARN: workstep #%d has multiple %ss\n",
		     ws-> entity_id(), name);
	}
	*mfp = fp;
    }
    /* unrecognized amr */
}

void link_amr_operation (
    action_method_relationship * amr,
    machining_operation * op
    )
{
    char * name = amr-> name();
    action_method * ed = amr-> related_method();

    if (!name || !*name) {
	if (1) //m_verbose) 
	    fprintf (stdout, "WARN: amr #%d has no name\n",
		     amr-> entity_id());
	return;
    }

    if (!strcmp (name, "technology"))
    {
	if (!ed-> isa (ROSE_DOMAIN(machining_technology)))
	{
	    if (1) //m_verbose) 
		fprintf (stdout, "ERR: amr #%d bad technology\n",
			 amr-> entity_id());
	    return; 
	}
	if (op-> m_tech && 1){ //m_verbose) {
	    fprintf (stdout, "WARN: operation #%d has multiple technologies\n",
		     op-> entity_id());
	}
	op-> m_tech = ROSE_CAST(machining_technology, ed);
    }

    else if (!strcmp (name, "machine functions"))
    {
	if (!ed-> isa (ROSE_DOMAIN(machining_functions)))
	{
	    if (1) //m_verbose) 
		fprintf (stdout, "ERR: amr #%d bad machining functions\n",
			 amr-> entity_id());
	    return; 
	}
	if (op-> m_funs && 1){ //m_verbose) {
	    fprintf (stdout, "WARN: operation #%d has multiple functions\n",
		     op-> entity_id());
	}
	op-> m_funs = ROSE_CAST(machining_functions, ed);
    }

    else if (!strcmp (name, "machining strategy"))
    {
	if (!ed-> isa (ROSE_DOMAIN(machining_strategy)))
	{
	    if (1) //m_verbose) 
		fprintf (stdout, "ERR: amr #%d bad machining strategy\n",
			 amr-> entity_id());
	    return; 
	}
	if (op-> m_strategy && 1){ //m_verbose) {
	    fprintf (stdout, "WARN: operation #%d has multiple strategies\n",
		     op-> entity_id());
	}
	op-> m_strategy = ROSE_CAST(machining_strategy, ed);

	/* Some additional type checks */
	if (1 && //m_verbose && 
	    op->  isa (ROSE_DOMAIN(drilling_type_operation)) &&
	    !ed-> isa (ROSE_DOMAIN(drilling_type_strategy))) {
	    fprintf (stdout, "WARN: drilling #%d: strategy #%d not drilling\n",
		     op-> entity_id(), ed-> entity_id());
	}
	if (1 && //m_verbose && 
	    op->  isa (ROSE_DOMAIN(milling_type_operation)) &&
	    !ed-> isa (ROSE_DOMAIN(milling_type_strategy))) {
	    fprintf (stdout, "WARN: milling #%d: strategy #%d not milling\n",
		     op-> entity_id(), ed-> entity_id());
	}
	if (1 && //m_verbose && 
	    op->  isa (ROSE_DOMAIN(freeform_milling_operation)) &&
	    !ed-> isa (ROSE_DOMAIN(freeform_milling_strategy))) {
	    fprintf (stdout, "WARN: freeform #%d: strategy #%d not freeform\n",
		     op-> entity_id(), ed-> entity_id());
	}
    }

    else if (!strcmp (name, "approach strategy") ||
	     !strcmp (name, "retract strategy"))
    {
	machining_approach_retract_strategy ** arp;

	if (!op-> isa (ROSE_DOMAIN(milling_type_operation)))
	{
	    if (1) //m_verbose) 
		fprintf (stdout, "ERR: amr #%d only milling ops have a/r\n",
			 amr-> entity_id());
	    return; 
	}

	if (!ed-> isa (ROSE_DOMAIN(machining_approach_retract_strategy)))
	{
	    if (1) //m_verbose) 
		fprintf (stdout, "ERR: amr #%d bad approach/retract\n",
			 amr-> entity_id());
	    return; 
	}

	arp = (!strcmp (name, "approach strategy")?
	       &(op->m_approach):
	       &(op->m_retract));


	if (*arp && 1){ //m_verbose) {
	    fprintf (stdout, "WARN: operation #%d has multiple a/r strats\n",
		     op-> entity_id());
	}

	*arp = ROSE_CAST(machining_approach_retract_strategy, ed);
    }

    // name = "toolpath"
    // type of sequential method
    // relating = machining_operation or 
    //      machining_approach_retract_strategy
    // related = machining_toolpath
}


void StixNC::link_amrs()
{
    RoseCursor objs;
    RoseObject * obj;

    /* Go through all sequential action methods AMRs and put in
     * backpointers to the workplans.  Once they are in place, sort 
     * them so that we get the contents in order.
     *
     * The real question is whether to sort them all at once and then 
     * assign them as backpointers, or to sort each workplan individually.
     */

    // name = "final feature"
    // name = "in-process feature"
    // relating = machining_feature_process
    // related = machining_workingstep

    // name = "operation"
    // relating = machining_workingstep
    // related = machining_operation

    // name = "technology"
    // relating = machining_operation or machining_toolpath
    // related = machining_technology

    // name = "machine functions"
    // relating = machining_operation or machining_toolpath
    // related = machining_functions

    // name = "machining strategy"
    // relating = machining_operation
    // related = machining_strategy
    // drilling_type_operation ==> drilling_type_strategy
    // freeform_milling_operation ==> freeform_milling_strategy
    // milling_type_operation ==> milling_type_strategy

    // name = "approach strategy"
    // name = "retract strategy"
    // relating = milling_type_operation
    // related = machining_approach_retract_strategy

    // name = "toolpath"
    // type of sequential method
    // relating = machining_operation or 
    //      machining_approach_retract_strategy
    // related = machining_toolpath

    // name = "process model"
    // type of sequential method
    // relating = machining_functions
    // related = machining_process_model

    // name = "true branch"	(if statement)
    // name = "false branch"	(if statement)
    // name = "branch"		(selective)
    // name = "body"		(while statement)
    // name = "concurrent element"	(nonsequential, parallel)
    // type of concurrent_action_method  (nonsequential, parallel)
    // relating = mpe 
    // related = mpe

    // name = "workplan element"
    // type of sequential method
    // relating = machining_workplan
    // related = mpe

    if (m_verbose) fprintf (stdout, "Linking action method relations\n");

    objs.traverse (m_design);
    objs.domain (ROSE_DOMAIN(action_method_relationship));
    while (obj=objs.next()) {
	action_method_relationship * amr = 
	    ROSE_CAST(action_method_relationship, obj);

	action_method * ing = amr-> relating_method();
	action_method * ed  = amr-> related_method();

	if (!ing || !ed) continue; /* bad data */

	/* OPERATIONS */
	if (ing-> isa(ROSE_DOMAIN(machining_operation))) {
	    link_amr_operation (amr, ROSE_CAST(machining_operation, ing));
	}

	/* WORKPLANS */
	else if (ing-> isa(ROSE_DOMAIN(machining_workplan))) {
	    link_amr_workplan (amr, ROSE_CAST(machining_workplan, ing));
	} 

	/* WORKINGSTEPS */
	else if (ing-> isa(ROSE_DOMAIN(machining_workingstep))) {
	    link_amr_workingstep (amr, ROSE_CAST(machining_workingstep, ing));
	}

	/* FEATURES */
	else if (ing-> isa(ROSE_DOMAIN(machining_feature_process))) {
	    link_amr_feature (amr, ROSE_CAST(machining_feature_process, ing));
	}
    }

    /* Sort the sequential methods hooked to the workplans */
    objs.domain (ROSE_DOMAIN(machining_workplan));
    while (obj=objs.next()) {
	machining_workplan * wp = ROSE_CAST(machining_workplan, obj);

	if (!wp->m_steps.size()) continue;

	/* sort the steps */
  	qsort ((void *) &(wp->m_steps[0]), wp->m_steps.size(),
  	   sizeof (void *), &sequential_method_cmp);
    }
}

void StixNC::link_process_assocs()
{
    RoseCursor objs;
    RoseObject * obj;

    if (m_verbose) fprintf (stdout, "Linking process associations\n");

    /* find the root workplan for a project */
    objs.traverse (m_design);
    objs.domain (ROSE_DOMAIN(process_product_association));
    while (obj=objs.next()) {
	RoseObject * pd_obj;
	char * name;
	action_method * am;
	product * p;
	product_definition * pd;
	process_product_association * ppa = 
	    ROSE_CAST(process_product_association, obj);

	/* look at the name of the action, and the types of the
	 * arguments.  This is used to encode several different
	 * relationships.
	 *
	 * 'workpiece complete probing' and 'workpiece probing' 
	 * link machining_touch_probing to a workpiece.
	 * 'machining' links machining_workplan to project
	 * 'setup'     links machining_workplan to setup
	 *
	 * 'setup' also links machining process executable 
	 * to a setup for operator instructions, although 
	 * this type will change from ballot resolution
	 */

	if (!ppa->process()) continue; /* malformed data */

	pd_obj = get_select_object(ppa->defined_product());
	if (!pd_obj ||
	    !pd_obj-> isa(ROSE_DOMAIN(product_definition)))
	{
	    continue; /* everything we care about is a pdef */
	}
	pd = ROSE_CAST(product_definition, pd_obj);
	p = (pd->formation()? pd->formation()->of_product(): 0);

	name = ppa->process()-> name();
	am =   ppa->process()-> chosen_method();
	if (!p || !am || !name || !*name) {
	    continue; /* should never happen */
	}

	if (!strcmp(name, "machining"))
	{
	    StixProject * idx = get_project_idx (p);
	    if (!idx || !(p->m_catflags & CAT_PROJECT))
		continue;

	    if (!am-> isa(ROSE_DOMAIN(machining_workplan)))
		continue;

	    idx-> m_workplan = ROSE_CAST(machining_workplan,am);
	    idx-> m_workplan_ppa = ppa;
	}
	else if (!strcmp(name, "setup"))
	{
	    /* both uses relate a setup */
	    if (!(p->m_catflags & CAT_SETUP))
		continue;

	    if (am-> isa(ROSE_DOMAIN(machining_workplan)))
	    {
		machining_workplan * wp = ROSE_CAST(machining_workplan,am);
		wp->m_setup = pd;  
		wp->m_setup_ppa = ppa;  
	    }
	    else {
		/* operator instructions on the setup */
	    }
	}
	else if (!strcmp(name, "workpiece probing") ||
		 !strcmp(name, "workpiece complete probing"))
	{
	    machining_touch_probing * tp;

	    if (!(p->m_catflags & CAT_WORKPIECE))
		continue;

	    if (!am-> isa(ROSE_DOMAIN(machining_touch_probing)))
		continue;

	    tp = ROSE_CAST(machining_touch_probing,am);
	    tp->m_workpiece = pd;  
	    tp->m_workpiece_ppa = ppa;  
	}
    }



    /* find the feature associations for a workingstep */
    objs.domain (ROSE_DOMAIN(process_property_association));
    while (obj=objs.next()) {
	RoseObject * sa_obj;
	char * name;
	action_method * am;
	shape_aspect * sa;
	machining_feature_process * mfp;
	process_property_association * ppa = 
	    ROSE_CAST(process_property_association, obj);

	/* This is only used to hook features to workingsteps, through
	 * a machining_feature_process action method.  The action that
	 * calls this out should have the name 'machining'.  AMR links
	 * between the MFP and the individual workingsteps determine
	 * whether it is the in-process or final feature.
	 */
	if (!ppa->process()) continue; /* malformed data */

	sa_obj = get_select_object(ppa->property_or_shape());
	if (!sa_obj ||
	    !sa_obj-> isa(ROSE_DOMAIN(shape_aspect)))
	{
	    continue; /* everything we care about is a pdef */
	}
	sa = ROSE_CAST(shape_aspect, sa_obj);

	name = ppa->process()-> name();
	am =   ppa->process()-> chosen_method();
	if (!am || !name || !*name) {
	    continue; /* should never happen */
	}

	if (!strcmp(name, "machining"))
	{
	    if (!am-> isa(ROSE_DOMAIN(machining_feature_process)))
		continue;   /* err, need mfp */

	    mfp = ROSE_CAST(machining_feature_process,am);
	    mfp-> m_feature = sa;
	}
	else {
	    /* unknown action */
	}
    }
}





void StixNC::link_pdrs()
{
    RoseCursor objs;
    RoseObject * obj;

    if (m_verbose) fprintf (stdout, "Linking pdef relationships\n");

    /* Scan the make_from_usage_options and stock to workpiece */
    objs.traverse (m_design);
    objs.domain (ROSE_DOMAIN(product_definition_relationship));
    while (obj=objs.next()) 
    {
	product_definition_relationship * pdr =
	    ROSE_CAST(product_definition_relationship, obj);

	product_definition * ing = pdr-> relating_product_definition(); 
	product_definition * ed  = pdr-> related_product_definition();
	product * p;

	if (!ing || !ed) continue; /* bad data */

	/* PDRs are only used in three places, 
	 * 1) workpiece to project, with name="workpiece"
	 * 2) workpiece to the setup, also with name="workpiece" and 
	 * 3) workpiece to workpiece, as stock, make_from_usage subtype
	 */

	/* check for rawpiece of workpiece */
	if (obj->isa (ROSE_DOMAIN(make_from_usage_option)))
	{
	    StixWorkpiece * idx = get_workpiece_idx (ing);
	    if (!idx) continue; /* not a workpiece */

	    idx-> m_stock = ed;
	}

	/* only interested in pdrs named workpiece right now */
	if (strcmp (pdr->name(), "workpiece"))
	    continue;

	/* get the product of the ing */
	p = (ing->formation()? ing->formation()->of_product(): 0);
	if (!p) continue; /* should never happen */

	/* relating a project */
	if (p->m_catflags & CAT_PROJECT) {
	    StixProject * idx = get_project_idx(p);
	    if (!idx) continue; /* should warn */
	    if (idx-> m_workpiece_pdr)
		fprintf (stdout, "WARN: project #%d has multiple workpieces\n",
			 p-> entity_id());
	    
	    idx-> m_workpiece_pdr = pdr;
	}
	
	/* relating a setup */
	if (p->m_catflags & CAT_SETUP) {
	    StixSetup * idx = get_setup_idx(p);
	    if (!idx) continue; /* should warn */
	    pdr-> m_next_pdr = idx-> m_workpiece_pdrs;
	    idx-> m_workpiece_pdrs = pdr;
	}
    }
}


void StixNC::link_materials()
{
    RoseCursor objs;
    RoseObject * obj;
    unsigned i, sz;

    if (m_verbose) fprintf (stdout, "Linking materials\n");

    /* Scan the make_from_usage_options and stock to workpiece */
    objs.traverse (m_design);
    objs.domain (ROSE_DOMAIN(material_designation));
    while (obj=objs.next()) 
    {
	material_designation * md = ROSE_CAST(material_designation, obj);
	SetOfcharacterized_definition * defs = md-> definitions();

	if (!defs) continue;
	for (i=0, sz=defs->size(); i<sz; i++) {
	    RoseObject * d = get_select_object (defs->get(i));
	    StixWorkpiece * idx;

	    if (!d-> isa(ROSE_DOMAIN(product_definition)))
		continue;
	    
	    idx = get_workpiece_idx(ROSE_CAST(product_definition,d));
	    if (idx) idx-> m_md = md;
	}
    }

    objs.domain (ROSE_DOMAIN(applied_document_reference));
    while (obj=objs.next()) 
    {
	applied_document_reference * adr = 
	    ROSE_CAST(applied_document_reference, obj);

	SetOfdocument_reference_item * items = adr-> items();
	if (!items) continue;

	for (i=0, sz=items->size(); i<sz; i++) {
	    RoseObject * d = get_select_object (items->get(i));

	    if (!d-> isa(ROSE_DOMAIN(material_designation)))
		continue;

	    material_designation * md = 
		ROSE_CAST(material_designation, d);

	    md-> m_doc = adr;
	}
    }
}

void StixNC::link_feature_final (machining_workingstep * ws)
{
    /* if an inproc feature is used, tell it about the final feature
     * that it contributes to */
    shape_aspect * ff = get_workingstep_final_feature(ws);
    shape_aspect * pf = get_workingstep_inproc_feature(ws);
    StixFeature * idx = get_feature_idx(pf);

    if (!idx) return;
    if (!idx-> m_final_sa) {
	idx-> m_final_sa = ff;
	return;
    }

    /* already set, check for mismatches */
    if (idx-> m_final_sa != ff)
	fprintf (stdout, "WARN: feature #%d is inproc for multiple final features\n",
		 pf-> entity_id());
}


void StixNC::workpiece_calc_extra()
{
    unsigned i, sz;
    
    /* WORKPIECES - All of the relationships should be linked in at
     * this point, so now we can loop over our previously built indicies
     * and compute any extra stuff. 
     */
    if (m_verbose) fprintf (stdout, "Building workpiece indicies\n");

    for (i=0, sz=m_wkpcs.size(); i<sz; i++) {
	StixWorkpiece * idx = (StixWorkpiece *) m_wkpcs[i];

	/*  find a bounding box around the stock geometry */
	/* Both the stock limits and the material are done from
	   the stock, */
	idx-> m_limits = calc_shape_limits (idx-> m_stock);
	idx-> m_bblimits = calc_bbox_limits (idx-> m_stock);
	idx-> m_material = calc_material (idx-> m_stock);
    }
}



void StixNC::workstep_calc_extra()
{
    unsigned i, sz;
    
    /* Workplans and Workingsteps - All of the relationships should be
     * linked in at this point, so now we can loop over our previously
     * built indicies and compute any extra stuff. 
     */
    if (m_verbose) fprintf (stdout, "Building workplan indicies\n");

    for (i=0, sz=m_execs.size(); i<sz; i++) {
	machining_process_executable * mpe = 
	    (machining_process_executable *) m_execs[i];

	/* tag with an enum to simplify later processing */
	mpe-> m_type = calc_exec_base_type (mpe);

	/* annotate the features with final/in-proc information.  This
	 * is because things like display want to know what the final
	 * feature a thing contributes to is.
	 */
	if (mpe-> isa(ROSE_DOMAIN(machining_workingstep)))
	    link_feature_final (ROSE_CAST(machining_workingstep,mpe));

	if (mpe-> isa(ROSE_DOMAIN(machining_workplan)))
	    calc_exec_all_worksteps (ROSE_CAST(machining_workplan,mpe));
    }
}


void StixNC::link_sars()
{
    RoseCursor objs;
    RoseObject * obj;

    if (m_verbose) fprintf (stdout, "Linking feature components\n");

    /* Traverse the shape defining relationships and put in
     * backpointers to the feature that they apply. 
     */
    objs.traverse (m_design);
    objs.domain (ROSE_DOMAIN(shape_defining_relationship));
    while (obj=objs.next()) {
	link_feature_sar(ROSE_CAST(shape_aspect_relationship,obj));
    }

    objs.domain (ROSE_DOMAIN(feature_component_relationship));
    while (obj=objs.next()) {
	link_feature_sar(ROSE_CAST(shape_aspect_relationship,obj));
    }

    /* FIXME -- for broken boss relationships.  Remove 
    *  once the code is corrected 
    */
    objs.exact (ROSE_DOMAIN(shape_aspect_relationship));
    while (obj=objs.next()) {
	link_direct_sar(ROSE_CAST(shape_aspect_relationship,obj));
    }
}

void StixNC::link_direct_sar (shape_aspect_relationship *sar)
{
    /* FIXME -- This should not normally happen, but bosses are 
     * being written out incorrectly.  Remove this once the software
     * is corrected.
     */
    shape_aspect * usage = sar->related_shape_aspect();
    StixFeature * idx;

    if (!usage) return;
    if (IS_FEATURE(usage)) {
	idx = get_feature_idx (usage);
	idx-> m_direct_sars.append(sar);
    }
}


void StixNC::link_feature_sar (shape_aspect_relationship *sar)
{
    /* for shape_defining_rels and feature_component_rels
    /* Interestingly, the linkage to the feature is on the ED side
     * of this relation (through another shape aspect) and the
     * thing that describes the component is the ING.  Based on
     * past experience, you would expect it to be the opposite.
     * The only rationale is to think of the profile/bottom/etc as
     * being a first class thing, and the SA that relates to the
     * feature is a second class usage object, hence the ED.
     */

    /* See what this points to, it should be a feature or else we
     * are in trouble.
     */
    RoseObject * fobj;
    shape_aspect * usage = sar->related_shape_aspect();
    shape_aspect * feat;
    StixFeature * idx;

    if (!usage || !usage-> of_shape()) {
	/* bad SDR */
	return;
    }
    
    fobj = get_select_object (usage->of_shape()->definition());
    if (!fobj || !fobj-> isa (ROSE_DOMAIN(shape_aspect))) {
	/* not pointing at a feature */
	printf ("WARN: sar #%d does not refer to a shape aspect\n",
		sar-> entity_id());
	return;
    }
    feat = ROSE_CAST (shape_aspect, fobj);

    if (!IS_FEATURE(feat)) {
	printf ("WARN: sar #%d does not refer to a feature\n",
		sar-> entity_id());
	return;	/* not pointing at a feature */
    }

    idx = get_feature_idx (feat);
    idx-> m_component_sars.append(sar);
}



void StixNC::feature_calc_extra()
{
    unsigned i, sz;

    /* Like the product definitions, shape aspects are used for many
     * different purposes.
     */
    if (m_verbose) fprintf (stdout, "Linking feature components\n");

    /* Refine the pockets and profiles based on additional information */
    for (i=0, sz=m_features.size(); i<sz; i++) {
	StixFeature * idx = (StixFeature*) m_features[i]; 
	shape_aspect * sa = idx-> m_sa;

	if (sa-> m_type == FEAT_POCKET)
	{
	    /* refine the pocket depending on the profile */
	    shape_aspect * prof = get_feature_component
		(sa, "profile usage");

	    if (!prof) continue;
	    if (IS_OPEN_PROFILE(prof)) {
		sa-> m_type = FEAT_POCKET_OPEN;
	    }
	    if (IS_CLOSED_PROFILE(prof)) {
		sa-> m_type = FEAT_POCKET_CLOSED;
	    }
	    continue;
	}

	if (sa-> m_type == FEAT_PROFILE)
	{
	    /* refine the tag depending on the type of the components
	     * associated with it.
	     */
	    /* general outside profile */
	    if (get_feature_component 
		(sa, "profile usage", "boundary occurrence")) {
		sa-> m_type = FEAT_PROF_OUTSIDE;
		continue;
	    }

	    /* Some type of shape profile */
	    if (get_feature_component 
		(sa, "profile floor usage")) {
		sa-> m_type = FEAT_PROF_SHAPE;
	    }

	    /* perhaps refine further */
	    if (get_feature_component 
		(sa, "profile floor usage", 
		 "closed rectangular boundary occurrence")) {
		sa-> m_type = FEAT_PROF_SHAPE_RECT;
		continue;
	    }
	}
    }
}


/* ---------------------------------------- */
/* FIND AIM FROM EXTERNAL IDENTIFIER */
/* ---------------------------------------- */

product_definition * StixNC::exid_to_workpiece (unsigned id)
{
    StixWorkpiece * idx = exid_to_workpiece_idx(id);
    return (idx? idx->m_pdef: 0);
}

shape_aspect * StixNC::exid_to_feature (unsigned id)
{
    StixFeature * idx = exid_to_feature_idx(id);
    return (idx? idx->m_sa: 0);
}

shape_aspect * StixNC::exid_to_component (unsigned id)
{
    StixComponent * idx = exid_to_component_idx(id);
    return (idx? idx->m_sa: 0);
}

product * StixNC::exid_to_project (unsigned exid)
{
    StixProject * idx = exid_to_project_idx(exid);
    return (idx? idx->m_product: 0);
}

product * StixNC::exid_to_setup (unsigned exid)
{
    StixSetup * idx = exid_to_setup_idx(exid);
    return (idx? idx->m_product: 0);
}


machining_process_executable * StixNC::exid_to_workstep (unsigned exid)
{
    /* the ids start at 1, so make sure to decrement! */
    exid = exid & EXID_BODYMASK; /* strip type info */
    return  ((exid && exid <= m_execs.size())?
	     ((machining_process_executable *) m_execs[exid-1]): 0);
}


machining_tool * StixNC::exid_to_tool (unsigned exid)
{
    /* the ids start at 1, so make sure to decrement! */
    exid = exid & EXID_BODYMASK; /* strip type info */
    return  ((exid && exid <= m_tools.size())?
	     ((machining_tool *) m_tools[exid-1]): 0);
}


/* ---------------------------------------- */
/* FIND INDEX DATA FROM EXTERNAL IDENTIFIER */
/* ---------------------------------------- */

StixProject * StixNC::exid_to_project_idx (unsigned id)
{
    /* the ids start at 1, so make sure to decrement! */
    id = id & EXID_BODYMASK; /* strip type info */
    return  ((id && id <= m_projects.size())?
	     ((StixProject*) m_projects[id-1]): 0);
}

StixSetup * StixNC::exid_to_setup_idx (unsigned id)
{
    /* the ids start at 1, so make sure to decrement! */
    id = id & EXID_BODYMASK; /* strip type info */
    return  ((id && id <= m_setups.size())?
	     ((StixSetup*) m_setups[id-1]): 0);
}

StixWorkpiece * StixNC::exid_to_workpiece_idx (unsigned id)
{
    /* the ids start at 1, so make sure to decrement! */
    id = id & EXID_BODYMASK; /* strip type info */
    return  ((id && id <= m_wkpcs.size())?
	     ((StixWorkpiece*) m_wkpcs[id-1]): 0);
}

StixFeature * StixNC::exid_to_feature_idx (unsigned id)
{
    /* the ids start at 1, so make sure to decrement! */
    id = id & EXID_BODYMASK; /* strip type info */
    return  ((id && id <= m_features.size())?
	     ((StixFeature*) m_features[id-1]): 0);
}

StixComponent * StixNC::exid_to_component_idx (unsigned id)
{
    /* the ids start at 1, so make sure to decrement! */
    id = id & EXID_BODYMASK; /* strip type info */
    return  ((id && id <= m_comps.size())?
	     ((StixComponent*) m_comps[id-1]): 0);
}




product * StixNC::default_project()
{
    if (m_projects.size()) {
	StixProject * idx = (StixProject*) m_projects[0];
	return idx->m_product;
    }
    return 0;
}

product_definition * StixNC::default_workpiece()
{
    unsigned i, sz;
    /* we might call this on other APs, in which case we will not
     * have any projects, but we might have a workpiece. If there 
     * is no project, return the first non stock workpiece. 
     */

    product* p = default_project();
    if (p) return get_project_workpiece(p);

    /* look for a workpiece find the first one that is not a
     * stockpiece of something else */

    ROSE.beginTraversal();
    for (i=0, sz=m_wkpcs.size(); i<sz; i++) {
	StixWorkpiece * idx = (StixWorkpiece*) m_wkpcs[i];
	if (idx->m_stock) idx->m_stock->visit();
    }
    ROSE.endTraversal();

    for (i=0, sz=m_wkpcs.size(); i<sz; i++) {
	StixWorkpiece * idx = (StixWorkpiece*) m_wkpcs[i];
	if (!idx->m_pdef-> wasVisited()) return idx->m_pdef;
    }
    return 0;
}



/*======================================================================*/
/*======================================================================*/
/*======================================================================*/
/* FUNCTIONS FOR INDEX STRUCTURES */


StixWorkpiece::~StixWorkpiece()
{
    /* calculated values */
    if (m_limits) 	delete m_limits;
    if (m_bblimits) 	delete m_bblimits;

    if (m_material) 	delete m_material;
}


StixMaterial::~StixMaterial()
{
    unsigned i, sz;

    /* Delete the properties */
    for (i=0, sz=m_mtl_props.size(); i<sz; i++) {
	StixProperty * prop = (StixProperty*) m_mtl_props[i]; 
	delete prop;
    }    
    m_mtl_props.empty();
}
