/* $RCSfile: show_feature.cxx,v $
 * $Revision: 1.3 $ $Date: 2003/05/09 19:45:01 $
 * Auth: Dave Loffredo (loffredo@steptools.com)
 *
 * Sample program to print out some information about the 
 * features in a file.  This shows how to:
 *
 * 	- load the StixNC index with a data set.
 * 	- loop over all of the features
 *	- switch() on the type of features
 * 	- get some feature properties
 *	- switch() on the type of feature components
 */

#include <stdio.h>
#include <string.h>

#include <rose.h>
#include <integrated_cnc_schema.h>
#include <integrated_cnc_schema_ROSE_LOAD.h>
#include <stix.h>

void show_all_features (StixNC * stepnc);
void show_feature   (shape_aspect * feat);
void show_component (shape_aspect * comp);
void show_location  (axis2_placement_3d * ap);

/* some geometry info for general profiles */
void show_shape_path (shape_representation * rep);
void show_oriented_edge (oriented_edge * oe);
void show_vertex (vertex * v);

static void usage (const char * name)
{
    fprintf (stderr, "%s: <filename>\n", name);
    exit (1);
}


int main (int argc, char ** argv)
{
    StixNC stepnc;

    /* Force linkers to bring in all C++ classes.  Without this, some
     * linkers leave out classes and we get "best-fit" messages when
     * matching EXPRESS to the C++ classes.  
     */
//    ROSE.quiet(1);
    integrated_cnc_schema_force_load();

    /* must have at least one arg */
    if (argc != 2) usage (argv[0]);

    /* load the data set */
    stepnc.load (argv[1]);
    if (!stepnc.design()) {
	printf ("Design %s not found.\n", argv[1]);
	exit (1);
    }

    printf ("Examining Design: %s\n", argv[1]);
    show_all_features(&stepnc);
    return 0;
}


/* Print out some basic information on the contents of the file */
void show_all_features(StixNC * stepnc)
{
    printf ("\n------------------------------\n");
    printf ("Features: ");

    StixFeatureCursor features (stepnc);
    shape_aspect * feat;

    printf ("%d present\n", features.size());
    while (feat = features.next()) {
	show_feature (feat);
    }
}


void show_feature (shape_aspect * feat)
{
    shape_aspect * sa;
    StixFeatureBossCursor bosses (feat);

    /* print the step file id */
    printf ("-----\n");
    printf ("Instance #%d: ", feat-> entity_id());

    /* look at the type of feature */
    switch (get_aspect_type (feat)) {
    case FEAT_BOSS:
	printf ("boss (ignored)\n");
	return; /* ignore bosses, not a primary features */

    case FEAT_HOLE:
	printf ("Round Hole Feature\n");
	show_location (get_feature_placement(feat));
	printf ("hole start diameter: %.15g\n",
		get_hole_diameter (feat));

	printf ("hole depth: %.15g\n",
		get_path_linear_distance(get_hole_path (feat)));

	/* print the taper if present */
	sa = get_hole_taper(feat);
	if (sa) show_component(sa);

	/* print the bottom condition if present */
	sa = get_hole_bottom_condition(feat);
	if (sa) show_component(sa);
	break;

    case FEAT_PLANE:
	printf ("Planar Face Feature\n");
	/* print the calculated feature depth by intersecting 
	 * the feature profile with the stock geometry.
	 */
	show_location (get_feature_placement(feat));
	printf ("planar face depth (calculated): %.15g \n", 
		calc_feature_depth(feat));

	if (bosses.size()) {
	    printf ("bosses: ");
	    while (sa = bosses.next()) 
		printf ("#%d ", sa->entity_id());
	    printf ("\n");
	}
	break;

    case FEAT_POCKET:
    case FEAT_POCKET_OPEN:
    case FEAT_POCKET_CLOSED:
	printf ("Pocket Feature\n");
	show_location (get_feature_placement(feat));

	printf ("pocket depth: %.15g\n",
		get_path_linear_distance
		(get_pocket_path (feat)));

	printf ("pocket boundary: ");
	show_component(get_pocket_profile(feat));

	/* print the taper if present */
	sa = get_pocket_taper(feat);
	if (sa) show_component(sa);

	if (bosses.size()) {
	    printf ("bosses: ");
	    while (sa = bosses.next()) 
		printf ("#%d ", sa->entity_id());
	    printf ("\n");
	}
	break;


    case FEAT_STEP:
	printf ("Step Feature\n");

	/* print the calculated feature depth by intersecting 
	 * the feature profile with the stock geometry.
	 */
	show_location (get_feature_placement(feat));
	printf ("step depth (calculated): %.15g \n", 
		calc_feature_depth(feat));

	if (bosses.size()) {
	    printf ("bosses: ");
	    while (sa = bosses.next()) 
		printf ("#%d ", sa->entity_id());
	    printf ("\n");
	}
	break;


    case FEAT_PROF_OUTSIDE:
	printf ("Outside Profile\n");
	show_location (get_feature_placement(feat));

	printf ("profile depth: %.15g\n",
		get_path_linear_distance
		(get_outside_profile_sweep_path (feat)));

	printf ("profile boundary: ");
	show_component(get_outside_profile_feature_boundary(feat));

	break;


    case FEAT_CBORE:
    case FEAT_CHAMFER:
    case FEAT_COMPOUND:
    case FEAT_CSUNK:
    case FEAT_EDGE_ROUND:
    case FEAT_PAT_CIRC:
    case FEAT_PAT_GEN:
    case FEAT_PAT_RECT:
    case FEAT_PROFILE:
    case FEAT_PROF_SHAPE:
    case FEAT_PROF_SHAPE_RECT:
    case FEAT_REGION:
    case FEAT_ROUNDED_END:
    case FEAT_SLOT:
    case FEAT_SPHERICAL_CAP:
    case FEAT_THREAD:
    case FEAT_TOOLPATH:
    default:
	/* other types of features */
	printf ("Feature of type: %s\n", get_aspect_typename (feat));
	break;
    }
}


void show_location (axis2_placement_3d * ap)
{
    /* This shows the raw AP-238 feature placement.  Note that the
     * feature placements in ISO 14649 are not quite harmonized with
     * the STEP APs.  Look at the calc_feature_orientation() function 
     * for calculations to transform to ISO 14649 conventions.
     */
    ListOfDouble * vals;
    unsigned i;

    printf ("placement and orientation: (#%d)\n", 
	    (ap? ap-> entity_id(): 0));

    if (!ap) {
	printf ("\tno placement!\n");
	return;
    }

    /* For robustness, these should really be testing for 
     * null objects rather than just chaining together the
     * access functions
     */

    /* print the orientation axes */ 
    vals = ap->ref_direction()->direction_ratios();
    printf ("\tref direction (x):  %.15g %.15g %.15g\n",
	    vals-> get(0), vals-> get(1), vals-> get(2));

    vals = ap->axis()->direction_ratios();
    printf ("\taxis direction (z): %.15g %.15g %.15g\n",
	    vals-> get(0), vals-> get(1), vals-> get(2));

    /* print the location of the origin */ 
    vals = ap->location()->coordinates();
    printf ("\torigin location:    %.15g %.15g %.15g\n",
	    vals-> get(0), vals-> get(1), vals-> get(2));
}



void show_component (shape_aspect * comp)
{
    /* print out details on the profiles, paths, tapers,
     * and bottom conditions associated with features
     */

    /* look at the type of feature */
    switch (get_aspect_type (comp)) {

	/* simple open profiles */
    case PROF_LINEAR:
	printf ("linear profile of length: %.15g\n", 
		get_profile_linear_length(comp));
	return;

    case PROF_OPEN_GENERAL:
	printf ("open general profile with shape path:\n");
	show_shape_path (get_profile_general_shape (comp));
	return;

    case PROF_SQUARE_U:
	printf ("square u profile %.15g wide\n", 
		get_profile_square_u_width(comp));
	printf ("\tangle one %.15g, radius %.15g\n", 
		get_profile_square_u_first_angle (comp),
		get_profile_square_u_first_radius (comp));
	printf ("\tangle two %.15g, radius %.15g\n", 
		get_profile_square_u_second_angle (comp),
		get_profile_square_u_second_radius (comp));
	return;

    case PROF_OPEN_CIRC:
    case PROF_ROUNDED_U:
    case PROF_TEE:
    case PROF_VEE:
	printf ("open profile of type: %s\n", 
		get_aspect_typename(comp));
	return;
	    

	/* simple closed profiles */
    case PROF_CLOSED_RECT:
	printf ("rectangular profile %.15g by %.15g \n", 
		get_profile_rectangular_length(comp),
		get_profile_rectangular_width(comp));
	return;
    case PROF_CLOSED_CIRC:
	printf ("circular profile diameter %.15g \n", 
		get_profile_circle_diameter(comp));
	return;

    case PROF_CLOSED_GENERAL:
	printf ("closed general profile with shape path:\n");
	show_shape_path (get_profile_general_shape (comp));
	return;

    case PROF_CLOSED_NGON:
	printf ("closed profile of type: %s\n", 
		get_aspect_typename(comp));
	return;

	/* paths */
    case PATH_LINEAR:
    case PATH_OPEN_CIRC:
    case PATH_CLOSED_CIRC:
    case PATH_GENERAL:

	/* hole bottom types */
    case HBOT_FLAT:
    case HBOT_FLAT_RADIUS:
    case HBOT_SPHERICAL:
    case HBOT_CONICAL:
    case HBOT_THROUGH:
	printf ("bottom condition of type: %s\n", 
		get_aspect_typename(comp));
	return;


	/* pocket bottom types */
    case PBOT_THROUGH:
    case PBOT_FLAT:
    case PBOT_RADIUSED:
    case PBOT_GENERAL:
	printf ("pocket bottom of type: %s\n", 
		get_aspect_typename(comp));
	return;


	/* slot end types */
    case SLOTEND_WOODRUFF:
    case SLOTEND_RADIUSED:
    case SLOTEND_FLAT:
    case SLOTEND_OPEN:
    case SLOTEND_LOOP:
	printf ("slot end of type: %s\n", 
		get_aspect_typename(comp));
	return;


	/* taper types */
    case TAPER_DIAMETER:
	printf ("taper to diameter: %.15g \n", 
		get_taper_final_diameter (comp));
	return;

    case TAPER_ANGLE:
	printf ("taper angle: %.15g \n", 
		get_taper_angle (comp));
	return;

	/* profile floors */
    case PROFLOOR_THROUGH:
    case PROFLOOR_PLANAR:
    case PROFLOOR_GENERAL:
    default:
	break;
    }
}



/* ==================================================
 * Print Some Geometry Information --
 * 
 * General profiles and general paths are represented by some 
 * geometric information.  When we see one of those in a feature
 * description, we want to print a brief summary of the geometry.
 */
void show_shape_path (shape_representation * rep)
{
    unsigned i, sz;
    if (!rep) {
	printf ("\tno shape representation!\n");
	return;
    }
	
    representation_item * pri = get_representation_item
	(rep, ROSE_DOMAIN(path));

    path * p = (pri? ROSE_CAST(path,pri):0);
    
    if (!p || !p-> edge_list()) {
	printf ("\tNo shape path!\n");
	return;
    }

    printf ("\tShape path has %d edges\n", p-> edge_list()-> size());
    for (i=0, sz=p-> edge_list()-> size(); i<sz; i++) {
	printf ("\t[%d] ", i);
	show_oriented_edge (p-> edge_list()-> get(i));
    }
}

    
/* Print some basic details of each oriented_edge, like the curve
 * type, start & end points, etc.
 */
void show_oriented_edge (oriented_edge * oe)
{
    if (!oe) return;

    edge * e = (oe? oe-> edge_element(): 0);
    edge_curve * ec = 
	((e && e-> isa(ROSE_DOMAIN(edge_curve)))? 
	 ROSE_CAST(edge_curve,e): 0);
    curve * c = (ec? ec-> edge_geometry(): 0);

    if (!c) {
    	printf ("No edge curve!\n");
	return;
    }

    printf ("edge type: %s\n", c-> domain()->name());

    printf ("\torientation: %s %s\n",
	    (oe-> orientation()? "edge=norm":  "edge=flip"),
	    (ec-> same_sense()?  "curve=same": "curve=flip"));

    printf ("\tstart: "); show_vertex (e-> edge_start());
    printf ("\tend:   "); show_vertex (e-> edge_end());
}


void show_vertex (vertex * v)
{
    unsigned i, sz;
    vertex_point * vp =
	((v && v->isa(ROSE_DOMAIN(vertex_point)))?
	 ROSE_CAST(vertex_point,v) : 0);

    point * p = (vp? vp-> vertex_geometry(): 0);
    cartesian_point * cp = 
	((p && p->isa(ROSE_DOMAIN(cartesian_point)))?
	 ROSE_CAST(cartesian_point,p) : 0);

    ListOfdouble * coords = (cp? cp-> coordinates(): 0);

    if (!coords) 
	printf ("<null vertex>\n");
    else
	printf ("%.15g %.15g %.15g\n",
		coords-> get(0), 
		coords-> get(1), 
		coords-> get(2));
}
