/* $RCSfile: main.cxx,v $
 * $Revision: 1.15 $ $Date: 2003/08/12 16:05:17 $
 * Auth: Dave Loffredo (loffredo@steptools.com)
 * 
 * Copyright (c) 1991-2003 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * This file is part of the STIX Library
 * 
 * This file may be distributed and/or modified under the terms of 
 * the GNU General Public License version 2 as published by the Free
 * Software Foundation and appearing in the file LICENSE.GPL included
 * with this file.
 * 
 * THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTY OF ANY KIND,
 * INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE.
 * 
 * STEP Tools Inc. also offers licenses for commercial use of this
 * software in non-GPL applications.  Contact info@steptools.com for
 * terms and pricing.
 * 
 * 		----------------------------------------
 * 
 *  Test harness program for the AP-238 NC Interface
 */

#include <stdio.h>
#include <string.h>

#include <rose.h>
#include <integrated_cnc_schema.h>
#include <integrated_cnc_schema_ROSE_LOAD.h>
#include "stix.h"

void dump_report (StixNC * stixnc);


static void usage (char * name)
{
    fprintf (stderr, "%s: <filename>\n", name);
    fprintf (stderr, "Options:\n");
    fprintf (stderr, "  -units: dump known unit coversions\n");
    exit (1);
}


#define NEXT_ARG(i,argc,argv) ((i<argc)? argv[i++]: 0)

int main (int argc, char ** argv)
{
    int i=1;
    char * arg;
    StixNC stepnc;
  
    /* Force linkers to bring in all C++ classes.  Without this, some
     * linkers leave out classes and we get "best-fit" messages when
     * matching EXPRESS to the C++ classes.  
     */
//    ROSE.quiet(1);
    integrated_cnc_schema_force_load();

    /* must have at least one arg */
    if (argc < 2) usage (argv[0]);

    /* get remaining keyword arguments */
    while (arg = NEXT_ARG(i,argc,argv))
    {
	/* command line options */
	if (!strcmp (arg, "-units")) 
	{	
	    print_conversion_factors();
	    exit (1);
	}
	else 
	{
	    stepnc.load (arg);

	    dump_report(&stepnc);
	}
    }
    return 0;
}


/* Print out some basic information on the contents of the file */
void dump_report (StixNC * stixnc)
{
    unsigned count;

    printf ("STEP-NC Index Contents Report --------\n");
    if (!stixnc-> design()) {
	printf ("No Design Loaded, done.\n");
	return;
    }
    printf ("Design: %s\n", stixnc-> design()->name());

    /* All imported as AP-238, but print out information on the
     * original source schema */
    switch (stixnc-> src_schema()) {
    case StixSchema_ap238:
	printf ("Schema: AP-238\n"); 	break;

    case StixSchema_ap203:
	printf ("Schema: AP-238, imported from AP-203\n"); 	break;

    case StixSchema_ap214:
	printf ("Schema: AP-238, imported from AP-214\n"); 	break;

    case StixSchema_ap224:
	printf ("Schema: AP-238, imported from AP-224\n"); 	break;

    case StixSchema_other:
	printf ("Schema: %s, attempted to read as AP-238\n",
		stixnc-> src_schema_name()); break;
	
    default:
	printf ("SCHEMA NOT RECOGNIZED\n");
    }


    printf ("\n------------------------------\n");
    printf ("Projects: ");

    StixProjectCursor projects (stixnc);
    product * pj;

    count=0;
    printf ("%d present\n", projects.size());
    while (pj = projects.next()) {
	machining_workplan * wp = get_project_workplan (pj);

	printf ("[%d] name = %s (id #%d) \n", count++, 
		printable_name(pj), pj->entity_id());

	if (wp) {
	    printf ("\tworkplan = %s (id #%d)\n", 
		    printable_name(wp), wp->entity_id());
	}
    }

    /* basic sanity check */
    if (!count)
	printf ("NO PROJECT!  Root workplan kept by project!\n");
    else if (count > 1)
	printf ("MULTIPLE PROJECTS!  Should only be one!\n");
    


    printf ("\n------------------------------\n");
    printf ("Setups: ");

    StixSetupCursor setups (stixnc);
    product * su;

    count=0;
    printf ("%d present\n", setups.size());



    printf ("\n------------------------------\n");
    printf ("Workpieces: ");

    StixWorkpieceCursor workpieces (stixnc);
    product_definition * wkpc;

    count=0;
    printf ("%d present\n", workpieces.size());
    while (wkpc = workpieces.next()) {
	printf ("[%d] ", count++);

	dump_workpiece_detail (wkpc);
    }



    printf ("\n------------------------------\n");
    printf ("Root Workplan: ");

    machining_workplan * wp = get_project_workplan
	(stixnc->default_project());
    
    dump_workplan_detail (wp);
    dump_workplan_all_tools (wp);
    dump_workplan_all_feats (wp);



    printf ("\n------------------------------\n");
    printf ("Features: ");

    StixFeatureCursor features (stixnc);
    shape_aspect * feat;

    count=0;
    printf ("%d present\n", features.size());
    while (feat = features.next()) {
	printf ("[%d] ", count++);
	
	/* this should filter out bosses */
	dump_feature_detail (feat);
    }


    printf ("\n------------------------------\n");
    printf ("Tools: ");

    StixToolCursor tools (stixnc);
    machining_tool * mt;

    count=0;
    printf ("%d present\n", tools.size());
    while (mt = tools.next()) {
	printf ("[%d] ", count++);
	dump_tool_detail (mt);
    }


    printf ("\n------------------------------\n");
    printf ("Dimensional Tolerances: ");

    StixDimToleranceCursor dtols (stixnc);
    plus_minus_tolerance * dtol;

    count=0;
    printf ("%d present\n", dtols.size());
    while (dtol = dtols.next()) {
	RoseObject * f1 = get_tolerance_face (dtol);
	RoseObject * f2 = get_tolerance_other_face (dtol);
	
	printf ("[%d] ", count++);
	printf ("type = [%s]  detail = [%s] (#%d)\n", 
		get_tolerance_typename(dtol), 
		get_tolerance_name(dtol), 
		dtol-> entity_id());

	printf ("\tvalue = %g [ %+g / %+g ]\n", 
		get_tolerance_nominal(dtol), 
		get_tolerance_upper(dtol), 
		get_tolerance_lower(dtol));

	if (f1) printf ("\ttoleranced face #%d\n", f1-> entity_id());
	if (f2) printf ("\ttoleranced other face #%d\n", f2-> entity_id());
    }

    printf ("\n------------------------------\n");
    printf ("Geometric Tolerances: ");

    StixGeomToleranceCursor gtols (stixnc);
    geometric_tolerance * gtol;

    count=0;
    printf ("%d present\n", gtols.size());
    while (gtol = gtols.next()) {
	RoseObject * f1 = get_tolerance_face (gtol);

	printf ("[%d] ", count++);
	printf ("type = [%s]  detail = [%s] (#%d)\n", 
		get_tolerance_typename(gtol), 
		get_tolerance_name(gtol), 
		gtol-> entity_id());

	printf ("\tvalue = %g (precision %g)\n", 
		get_tolerance_nominal(gtol), 
		get_tolerance_precision(gtol));

	switch (get_tolerance_limit_condition(gtol)) {
	case limit_condition_maximum_material_condition:
	    printf ("\tlimit = maximum material condition\n");
	    break; 
	case limit_condition_least_material_condition:
	    printf ("\tlimit = least material condition\n");
	    break; 
	case limit_condition_regardless_of_feature_size:
	    printf ("\tlimit = regardless of feature size\n");
	    break; 
	case limit_condition_NULL:
	    break; /* no action */
	}

	if (f1) printf ("\ttoleranced face #%d\n", f1-> entity_id());

	datum * d;
	int prec = 0;

	while (d = get_tolerance_datum (gtol, prec++)) {
	    RoseObject * dface = get_datum_face (d);
	    char * dname = d-> identification();

	    if (prec==1) printf ("\tdatums = ");
	    if (prec>1) printf (", ");

	    printf ("%s (face #%d)", (dname? dname : "unknown"), 
		    (dface? dface-> entity_id(): 0));
	}
	if (prec>1) printf ("\n");
    }
}
