/* $RCSfile: schema.cxx,v $
 * $Revision: 1.2 $ $Date: 2003/08/12 22:18:21 $
 * Auth: Dave Loffredo (loffredo@steptools.com)
 * 
 * Copyright (c) 1991-2003 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * This file is part of the STIX Library
 * 
 * This file may be distributed and/or modified under the terms of 
 * the GNU General Public License version 2 as published by the Free
 * Software Foundation and appearing in the file LICENSE.GPL included
 * with this file.
 * 
 * THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTY OF ANY KIND,
 * INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE.
 * 
 * STEP Tools Inc. also offers licenses for commercial use of this
 * software in non-GPL applications.  Contact info@steptools.com for
 * terms and pricing.
 * 
 * 		----------------------------------------
 */

/* Code to handle importing files defined by schemas other than
 * ap238.  Right now we read everything as if it were an ap238 file 
 * and jettison any extra entities.  In the future, we may try to do 
 * something more sophisticated by reading with a merged schema and 
 * bringing over selected data (like old-style assignments).
 */

#include <rose.h>
#include <RoseP21Parser.h>
#include <RoseP21Writer.h>

#include "stix_imp.h"

#define AP238_SCHEMA_NAME	"integrated_cnc_schema"


/* READING THE PART 21 FILE
 * 
 * When reading the header section of a Part 21 file, this function
 * will be called for each schema name in the FILE_SCHEMA() object.
 * This function takes a design pointer and a schema name.  It adds
 * the merged schema to the design and passes back the orginal schema
 * from the file in the merged_schema_type variable.  Instead of this
 * variable, you can change this to whatever communication method is
 * appropriate for your application.
 *
 * A return value of ROSE_OK indicates successful completion while a
 * non-zero value indicates an error.
 */

RoseErrorCode stix_schema_read (
    RoseDesign * 	design, 
    char * 		sname
    )
{
    StixSchemaType typ;

    if (!strcmp (sname, "INTEGRATED_CNC_SCHEMA")) {
	typ = StixSchema_ap238;
    }
    else if (!strcmp (sname, "CONFIG_CONTROL_DESIGN")) {
	typ = StixSchema_ap203;
    }
    else if (!strcmp (sname, "FEATURE_BASED_PROCESS_PLANNING")) {
	typ = StixSchema_ap224;
    }

    /* just look at the start of the string because some
     * processors put an ASN/1 identifier at the end 
     */
    else if (!strncmp (sname, "AUTOMOTIVE_DESIGN", 17)) {
	typ = StixSchema_ap214;
    }
    else {
	typ = StixSchema_other;
    }

    /* Regardless of what we see, read using the ap238 schema.
     * Perhaps as we become more sophisticated, we may want to 
     * switch this to a merged 203/214/224/238 schema
     */
    RoseDesign * schema = stix_cnc_schema();
    if (!schema || !schema-> isSchema()) 
    {
	/* no fatal errors here */
	rose_io_ec()-> error ("Could not find %s schema file", AP238_SCHEMA_NAME);
	return ROSE_IO_BAD_SCHEMA;
    }
    design-> addSchema (schema);

    /* if a known AP type not 238, supress some errors */
    if (typ != StixSchema_ap238 && typ != StixSchema_other)
	stix_p21_quiet_mismatch_messages();

    /* record the originals for posterity */
    StixSchemaManager::put_src_schema(design, typ);
    StixSchemaManager::put_schema_name(design, sname);
    return ROSE_OK;
}



/*========================================
 * Functions to selectively disable certain error messages that would 
 * be an annoyance when reading mismatched schemas.
 */

void stix_p21_quiet_mismatch_messages()
{
    RoseError * err;
    err = rose_io_ec()-> find_error (ROSE_IO_BAD_DOMAIN);
    err-> f_severity = ROSE_SEV_NULL;  // was ROSE_SEV_WARNING

    /* print a message saying we are turning off the warnings? */
    rose_io_ec()-> warning ("Reading as AP-238, supressing 'unknown entity' messages.");

}

void stix_p21_restore_messages()
{
    /* restore all of the messages to normal severity */
    RoseError * err;
    err = rose_io_ec()-> find_error (ROSE_IO_BAD_DOMAIN);
    err-> f_severity = ROSE_SEV_WARNING;
}



RoseDesign * stix_cnc_schema()
{
    /* Regardless of what we see, read using the ap238 schema.
     * Perhaps as we become more sophisticated, we may want to 
     * switch this to a merged 203/214/224/238 schema
     */
    RoseDesign * schema = ROSE.findDesign(AP238_SCHEMA_NAME);
    if (!schema) return 0;

    /* add new shortname aliases to allow reading common types */
    DictionaryOfString * remap = schema-> aliases();

    if (!remap) {
	remap =  new DictionaryOfString;
	remap -> caseSensitive (ROSE_FALSE);
	schema-> aliases (remap);
    }

 
    /* Manually add some aliases to substitute a few types when
     * reading alternate schemas.  The right thing to do would be 
     * to read using a merged schema and then migrate the types
     * afterwards, but this way is pretty quick and easy.
     */

    /* only add them once */
    if (remap-> find ("AP-238 Import Mappings"))
	return schema;
    
    /* add (shortname/as-is , target/to-be) */
    remap-> add ("AP-238 Import Mappings", "present");

    /* AP-203 TYPES */
    /* not present in AP-238.  Will warn about the extra att */
    remap-> add ("product_definition_formation_with_specified_source",
		 "product_definition_formation");

    remap-> add ("cc_design_approval",
		 "applied_approval_assignment");

    remap-> add ("cc_design_date_and_time_assignment",
		 "applied_date_and_time_assignment");

    remap-> add ("cc_design_person_and_organization_assignment",
		 "applied_person_and_organization_assignment");

    remap-> add ("cc_design_security_classification",
		 "applied_security_classification_assignment");


    /* AP-224 TYPES */
    remap-> add ("feature_based_pp_approval_assignment",
		 "applied_approval_assignment");

    remap-> add ("feature_based_pp_person_and_organization_assignment",
		 "applied_person_and_organization_assignment");

    remap-> add ("feature_based_pp_security_classification_assignment",
		 "applied_security_classification_assignment");

    /* AP-224 uses date assignments rather than date_and_time.  We can
     * not trivially remap that since the attributes are of different
     * types.  We need to bring that over with a little semantic code 
     */

    return schema;
}


/*========================================
 * StixSchemaManager --  The schema manager and associated support 
 * functions let us read designs with different schemas.  This makes 
 * it a little easier to import different schemas.
 */

ROSE_IMPLEMENT_MANAGER_COMMON(StixSchemaManager)

StixSchemaManager * StixSchemaManager::find_manager (RoseDesign * obj)
{
    return (obj? ((StixSchemaManager *) obj-> find_manager(type())): 0);
}

StixSchemaManager * StixSchemaManager::make_manager (RoseDesign * obj)
{
    /* find or create the manager */
    if (!obj) return 0;
    StixSchemaManager * mgr = (StixSchemaManager *) obj-> find_manager (type());

    if (!mgr) {
	mgr = new StixSchemaManager;
	obj-> add_manager (mgr);
    }
    return mgr;
}

StixSchemaType StixSchemaManager::get_src_schema (RoseDesign * obj)
{
    StixSchemaManager * mgr = StixSchemaManager::find_manager(obj);
    return (mgr? mgr-> m_src: StixSchema_ap238);
}

void StixSchemaManager::put_src_schema (RoseDesign * obj, StixSchemaType ap)
{
    StixSchemaManager * mgr = StixSchemaManager::make_manager (obj);
    if (mgr) mgr-> m_src = ap;
}

char * StixSchemaManager::get_schema_name (RoseDesign * obj)
{
    StixSchemaManager * mgr = StixSchemaManager::find_manager(obj);
    return (mgr? mgr-> m_name: AP238_SCHEMA_NAME);
}

void StixSchemaManager::put_schema_name (RoseDesign * obj, const char * ap)
{
    StixSchemaManager * mgr = StixSchemaManager::make_manager (obj);
    if (mgr) mgr-> m_name = ap;
}


