/* $RCSfile: stix_imp.h,v $
 * $Revision: 1.24 $ $Date: 2003/08/12 22:18:21 $
 * Auth: Dave Loffredo (loffredo@steptools.com)
 * 
 * Copyright (c) 1991-2003 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * This file is part of the STIX Library
 * 
 * This file may be distributed and/or modified under the terms of 
 * the GNU General Public License version 2 as published by the Free
 * Software Foundation and appearing in the file LICENSE.GPL included
 * with this file.
 * 
 * THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTY OF ANY KIND,
 * INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE.
 * 
 * STEP Tools Inc. also offers licenses for commercial use of this
 * software in non-GPL applications.  Contact info@steptools.com for
 * terms and pricing.
 * 
 * 		----------------------------------------
 *
 * Private Index Structures 
 *
 * These contain the backpointers and such.  There are usually access
 * functions to go through these and get various bits.  You are better
 * off using those functions rather than accessing these directly.  
 *
 * I originally made these structs (now 100% public classes), so you 
 * can fiddle with them directly rather than using the functions, but
 * you WILL get yourself in trouble, and then I will laugh at you.
 */

#ifndef	STIX_IMP_H
#define	STIX_IMP_H

#include "stix.h"
#include "xform.h"

/* In general, I save pointers to the relationships rather than the
 * things they relate, this will make update easier in the future, and
 * for one to many relationships, I chain the rel objects together in
 * a single linked list with a next pointer.
 */

class idx_base {    /* base struct for all index records */
public:
    unsigned m_exid;	/* external identifier */

    idx_base() {
	m_exid=0;
    }
};    


/* This separate index object is needed because not all product
 * definitions are workpieces.  For the part_id mapping att, look at
 * m_shape.  For the stock_id, look at the m_shape of the stock.
 */
class StixWorkpiece : public idx_base {
public:
    product_definition * m_pdef;
    product_definition * m_stock;
    shape_representation * m_shape;
    shape_representation * m_bbox;	/* bounding box, sr with params */
    material_designation * m_md;

    StixShapeLimit * m_limits;
    StixShapeLimit * m_bblimits;
    StixMaterial * m_material;

    /* unit is computed */

    StixWorkpiece() {
	m_pdef=0;
	m_stock=0;
	m_shape=0;
	m_bbox=0;
	m_md=0;
	m_limits=0;
	m_bblimits=0;
	m_material=0;
    }

    ~StixWorkpiece();
};

/* This separate index object is needed because some shape aspects are
 * features, some are components and some are nothing.
 */
class StixFeature : public idx_base {
public:
    shape_aspect * m_sa;
    shape_aspect * m_final_sa;  /* for in-proc feats */

    /* features can have explicit representations with a bunch of
     * faces or other geometry in a shape rep, and some of them have 
     * an implicit representation as well, with params.
     */
    shape_representation * m_explicit;
    shape_representation * m_implicit;  /* shape rep with params */

    /* Only features are the target of shape aspect relationships,
     * with the possible exception of the bits in a replicate feature.
     * This contains shape definition and feature component rels, cast 
     * as type shape aspect relationship.
     *
     * Note that this keeps a list of those SARs that point to the 
     * feature through an intermediate "linking" shape aspect, thus:
     *
     * 		feat <- link_sa <- sar -> comp_sa
     *
     *  this does not contain any SARs that reference directly
     *
     *		feat <- sar -> other_sa
     */
    rose_vector		m_component_sars;

    /* this has the ones that link directly.  Only bosses do that, but
     * that is because of a software error.  They should be linking through 
     * an intermediate SA like all of the other component_sars above. 	
     * Delete this once the software is corrected.
     */
    rose_vector		m_direct_sars;

    StixFeature() {
	m_explicit=0;
	m_implicit=0;
	m_sa=0;
	m_final_sa=0;
    }
    /* no dtor needed */
};

class StixComponent : public idx_base {
public:
    shape_aspect * 		m_sa;

    /* components usually have implicit shape descriptions, 
     * but some may also have explicit ones 
     */
    shape_representation * m_explicit;
    shape_representation * m_implicit;  /* shape rep with params */
    
    StixComponent() {
	m_explicit=0;
	m_implicit=0;
	m_sa=0;
    }
    /* no dtor needed */
};

/* This separate index object is needed because some products are
 * projects, some are setups, and some are related to workpieces.
 */
class StixProject : public idx_base {
public:
    product * m_product;
    machining_workplan * 		m_workplan;
    process_product_association * 	m_workplan_ppa;
    product_definition_relationship*	m_workpiece_pdr;
    
    StixProject() {
	m_product=0;
	m_workplan=0;
	m_workplan_ppa=0;
	m_workpiece_pdr=0;
    }
    /* no dtor needed */
};

class StixSetup : public idx_base {
public:
    product * m_product;
    product_definition_relationship*	m_workpiece_pdrs;

    StixSetup() {
	m_product=0;
    }
    /* no dtor needed */
};


/* ======================================== */
/* CALCULATED VALUES */

/* These are not true index records, just a convenient place to
 * hold calculated values.  Since they will always be found via a
 * proper index, they do not have an exid.
 */
/* Lightweight cartesian point */
class StixPoint {
public:
    double m_x;
    double m_y;
    double m_z;

    StixPoint() {
	m_x=0; m_y=0; m_z=0;
    }
};

class StixShapeLimit {
public:
    StixPoint m_min;
    StixPoint m_max;
};

/* this is totally hacked up and did not work in the original
 * mappings, so we just keep the same constants.  When this is set to
 * work properly, we will point to the material designation and others.
 */
class StixMaterial {
public:
    RoseStringObject 	m_std_id;
    RoseStringObject 	m_mtl_id;
    rose_vector 	m_mtl_props;

    ~StixMaterial();
};

class StixProperty {
public:
    RoseStringObject m_name;
    RoseStringObject m_desc;
};


/* calulation functions */
unsigned calc_aspect_base_type (shape_aspect *);
unsigned calc_exec_base_type (machining_process_executable * mpe);

StixShapeLimit * calc_shape_limits (product_definition *);
StixShapeLimit * calc_bbox_limits (product_definition *);
StixMaterial *   calc_material (product_definition *);


/* We want to calculate these upon demand */
// Constructs a DFS listing of worksteps, stored in the workplan
// m_calc_all_steps attribute, while also detecting cycles.
//
int calc_exec_all_worksteps (
    machining_workplan * wp, 	/* wp to check */
    rose_stack * nested_wps = 0	/* use to detect cycles */
    );

int calc_exec_unique_features (
    machining_workplan * wp
    );

int calc_exec_unique_tools (
    machining_workplan * wp
    );

StixMtrx calc_feature_orientation(shape_aspect * feat);




/* ============================================================ */
/* Code to handle importing files defined by schemas other than
 * ap238.  Right now we read everything as if it were an ap238 file 
 * and jettison any extra entities.  In the future, we may try to do 
 * something more sophisticated by reading with a merged schema and 
 * bringing over selected data (like old-style assignments).
 */

/* hook functions to read/write the P21 schema name */
extern RoseErrorCode	stix_schema_read (RoseDesign *, char *);

/* read the AP-238 Schema */
RoseDesign * stix_cnc_schema();

/* functions to selectively disable certain error messages */
extern void stix_p21_quiet_mismatch_messages();
extern void stix_p21_restore_messages();


class StixSchemaManager : public RoseManager {
protected:
    StixSchemaType 	m_src;
    RoseStringObject 	m_name;

public:
    static StixSchemaManager * find_manager (RoseDesign *);
    static StixSchemaManager * make_manager (RoseDesign *);

    static StixSchemaType get_src_schema (RoseDesign *);
    static char * get_schema_name (RoseDesign *);

    static void put_src_schema (RoseDesign *, StixSchemaType);
    static void put_schema_name (RoseDesign *, const char *);

    StixSchemaManager() : m_src(StixSchema_other) { }

    ROSE_DECLARE_MANAGER_COMMON();
};


#endif /* STIX_IMP_H */
