/* $RCSfile: xform.h,v $
 * $Revision: 1.4 $ $Date: 2003/05/27 18:30:36 $
 * Auth: Ford Oxaal (ford@steptools.com)
 * Auth: Dave Loffredo (loffredo@steptools.com)
 * 
 * Copyright (c) 1991-2003 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * This file is part of the STIX Library
 * 
 * This file may be distributed and/or modified under the terms of 
 * the GNU General Public License version 2 as published by the Free
 * Software Foundation and appearing in the file LICENSE.GPL included
 * with this file.
 * 
 * THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTY OF ANY KIND,
 * INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE.
 * 
 * STEP Tools Inc. also offers licenses for commercial use of this
 * software in non-GPL applications.  Contact info@steptools.com for
 * terms and pricing.
 * 
 * 		----------------------------------------
 */

#ifndef XFORM_H
#define XFORM_H

/* In designing the StixMtrx and StixVctr classes, the basic two choices were:

   1. Whether to index by [row][column] or [column][row]
   -----------------------------------------------------
   StixMtrx is indexed by [x][y], == [position in row][position in
   column], == [column][row] -- THIS GOES AGAINST MATHEMATICAL
   CONVENTION, but is more intuitive in terms of viewing a matrix as
   embedded in an x,y Cartesian coordinate system.


   2. Whether matrices are composed of row or column vectors
   ---------------------------------------------------------
   StixMtrx (and trivially, StixVctr) are understood as composed of row
   vectors.


   Additionally, the following was considered:

   3. Whether matrix multiplication is defined by a. or b.
      a. (m x p) * (p x n) = (m x n)
      b. (p x m) * (n x p) = (n x m)
   ------------------------------------------------------------------
   We do not pick one or the other here.  Our StixVctr times StixMtrx (V * M)
   operation defined in StixVctr uses b., and StixMtrx times StixVctr (M * V)
   operation uses a.  Both (V * M) and (M * V) give the same result.


   4. Whether to begin indices with 1, or with 0.
   ----------------------------------------------
   StixMtrx and StixVctr indexing begins with 0.  THIS GOES AGAINST
   MATHEMATICAL CONVENTION, but is more intuitive in terms of computer
   programming.


   Implications of choices for matrix multiplication:
   --------------------------------------------------
   Matrix multiplication (*) is defined between two matrices where the
   first dimension of one matches the second dimension of the other.
   There are eight cases (permutations of the first 3 choices above)
   which determine two outcomes:

     a. whether matrix multiplication is defined by dotting the row(s)
     of the first with the corresponding column(s) of the second, or
     dotting the column(s) of the first with the corresponding row(s)
     of the second.

     b. whether a one-dimensional matrix (vector) is the first or
     second operand when multiplied by a two-dimensional matrix
     (matrix) (V * M or M * V).
   
   As noted above, StixMtrx and StixVctr are understood as row(s) of
   vector(s), and we use [column][row] indexing.  So we implement case 7. 
   For V * M (defined in StixVctr class) as well as M * M (defined in
   StixMtrx class) , and case 3. for M * V (defined in StixMtrx class).  

   The way StixMtrx and StixVctr work is somewhat intuitive: the thing you
   want to manipulate is the first operand, and the transform you want
   to apply is the second operand.  We implement M * V for the heck of
   it.

   1. (m x p) * (p x n) = (m x n)
      a. [row][column] indexing
         i. matrices understood as rows of vectors
	    row dot column
	    V * M
   2. (m x p) * (p x n) = (m x n)
      a. [row][column] indexing
         i. matrices understood as columns of vectors
	    row dot column
	    M * V
   3. (m x p) * (p x n) = (m x n)
      a. [column][row] indexing
         i. matrices understood as rows of vectors
	    column dot row
	    M * V
   4. (m x p) * (p x n) = (m x n)
      a. [column][row] indexing
         i. matrices understood as columns of vectors
	    column dot row
	    V * M
   5. (p x m) * (n x p) = (n x m)
      a. [row][column] indexing
         i. matrices understood as rows of vectors
	    column dot row
	    M * V
   6. (p x m) * (n x p) = (n x m)
      a. [row][column] indexing
         i. matrices understood as columns of vectors
	    column dot row
	    V * M
   7. (p x m) * (n x p) = (n x m)
      a. [column][row] indexing
         i. matrices understood as rows of vectors
	    row dot column
	    V * M
   8. (p x m) * (n x p) = (n x m)
      a. [column][row] indexing
         i. matrices understood as columns of vectors
	    row dot column
	    M * V

*/   

class StixVctr;

class StixMtrx {
private:
    double m[4][4];

public:
    /* Constructors */
    StixMtrx(); // identity matrix 
    StixMtrx(
	double a, double b, double c, double d,
	double e, double f, double g, double h,
	double i, double j, double k, double l,
	double m, double n, double o, double p
	);
  
    /* Operators */
    StixVctr operator * (StixVctr & V); // vector multiplication
    StixMtrx operator * (StixMtrx & M); // matrix multiplication
    StixMtrx operator * (double s); // scalar multiplication
    // note: assignment is default bit copy of array m 

    /* Members */
    StixVctr x();
    StixVctr y();
    StixVctr z();
    StixVctr w();
    void x(StixVctr &);
    void y(StixVctr &);
    void z(StixVctr &);
    void w(StixVctr &);
    double get(int, int);
    void put(double, int, int);
    StixMtrx t4(void); // transpose 4 x 4
    StixMtrx t3(void); // transpose 3 x 3 portion
    void normalizeInPlace();
    void leftHanded();

    double det2d(); // calculate the determinant for the 2d portion
    StixMtrx inverse(); // only for homogeneous matrices with orthonormal bases
    StixMtrx genTransformTo(StixMtrx & M); // only for orthogonal matrices

    // gen matrix for rotation about arbitrary axis
    void genRotation(StixVctr &, double radians); 
    void putTranslation(double a, double b, double c);

    // populate with three or four vectors
    int populate(StixVctr &, StixVctr &, StixVctr &);
    int populate(StixVctr &, StixVctr &, StixVctr &, StixVctr &); 
    void copy3(StixMtrx & M); // copy only 3 x 3 portion
    int isIdentity3();
    int isIdentity4();
    void print(FILE *, char * indent=0);
    void print_axes(FILE *, char * indent=0);
  
    /* Friends */
    /* move between cartesian transform operator and matrix */
    friend void StixMtrx_from_cto (RoseObject * cto, StixMtrx & M);
    friend void StixMtrx_to_cto (StixMtrx * M, RoseObject & cto);

    /* move between axis placements and matrix */
    friend void StixMtrx_from_ap (RoseObject * ax, StixMtrx & M);
    friend void StixMtrx_to_ap (StixMtrx & M, RoseObject * ax);
};


class StixVctr {
private:
    double v[4];

public:
    /* Constructors */
    StixVctr(void);
    StixVctr(double, double, double, double = 1);

    /* Operators */
    StixVctr operator * (StixVctr & w); // cross product
    StixVctr operator * (StixMtrx & m);     /* vector * matrix */

    /* vector * matrix -- implies a row vector is multiplied by a
     * matrix whose rows represent vectors.  A matrix * vector operation
     * is defined in the StixMtrx class, and implies a matrix whose columns
     * represent vectors, or , a transposed row matrix 
     */

    StixVctr operator * (double s); // scalar product
    StixVctr operator + (StixVctr & w); // scalar product
    StixVctr operator - (); // scalar product

    /* Members */
    double get(int i);
    void put(double val, int i);
    void normalizeInPlace(void); // in place normalization
    StixVctr normalize(void);
    StixVctr getDiffVec(void);
    StixVctr rightAngleTo(void);
    double length(void);
    int isZeroVec(double tol = 0.000001);
    void print(FILE *, char * indent = NULL);

    /* friends */
    friend void StixVctr_from_vec (RoseObject * vec, StixVctr & V); // load from STEP vector
    friend void StixVctr_from_dir (RoseObject * dir, StixVctr & V); // load from STEP direction
    friend void StixVctr_to_dir (StixVctr & V, RoseObject * dir);   // fill in STEP direction 
};

#endif  /* XFORM_H */
